use std::env;
use std::process::exit;

use sequoia_openpgp as openpgp;

use openpgp::Result;
use openpgp::policy::StandardPolicy;
use openpgp::parse::Parse;
use openpgp::cert::prelude::*;
use openpgp::types::HashAlgorithm;
use openpgp::types::RevocationStatus;

fn main() -> Result<()> {
    let args: Vec<String> = env::args().collect();

    if args.len() != 2 {
        eprintln!("Usage: {} KEYRING.pgp",
                  args.get(0).unwrap_or(&"sq-keyring-linter".into()));
        exit(2);
    }

    let filename = args.get(1).expect("checked above");

    // Standard policy.
    let sp = &StandardPolicy::new();
    // A standard policy that also accepts SHA1.
    let mut sha1 = StandardPolicy::new();
    sha1.accept_hash(HashAlgorithm::SHA1);
    let sha1 = &sha1;

    // Total number of certs.
    let mut certs = 0;

    // Certificates that are valid under the standard policy (this
    // means that they have a valid signature signature or a User ID
    // with a valid self signature).
    let mut certs_sp = 0;
    // Certificates that are valid under the standard policy + sha1
    // exception.
    let mut certs_sp_sha1 = 0;

    // Certificates that are valid and have a valid User ID.
    let mut certs_sp_userids = 0;
    let mut certs_sp_sha1_userids = 0;
    let mut certs_with_a_sha1_protected_userids = 0;
    let mut certs_with_only_sha1_protected_userids = 0;

    let mut certs_with_signing_subkeys = 0;
    let mut certs_with_sha1_protected_signing_subkeys = 0;
    let mut certs_with_sha1_protected_binding_sig = 0;
    let mut certs_with_sha1_protected_backsig = 0;

    let certp = CertParser::from_file(filename)?;
    for certo in certp {
        match certo {
            Ok(cert) => {
                certs += 1;

                let sp_vc = cert.with_policy(sp, None);
                if sp_vc.is_ok() {
                    certs_sp += 1;
                }
                let sp_sha1_vc = cert.with_policy(sha1, None);
                if sp_sha1_vc.is_ok() {
                    certs_sp_sha1 += 1;
                } else {
                    println!("Certificate {} is not valid under the standard policy + sha1",
                             cert.fingerprint());
                }
                if sp_vc.is_err() || sp_sha1_vc.is_err() {
                    continue;
                }
                let sp_vc = sp_vc.unwrap();
                let sp_sha1_vc = sp_sha1_vc.unwrap();

                let not_revoked = |ua: &ValidUserIDAmalgamation| -> bool {
                    if let RevocationStatus::Revoked(_)
                        = ua.revocation_status()
                    {
                        false
                    } else {
                        true
                    }
                };

                let sp_userids: Vec<_> = sp_vc
                    .userids()
                    .filter(not_revoked)
                    .map(|ua| ua.userid())
                    .collect();
                if sp_userids.len() > 0 {
                    certs_sp_userids += 1;
                }
                let sp_sha1_userids: Vec<_> = sp_sha1_vc
                    .userids()
                    .filter(not_revoked)
                    .map(|ua| ua.userid())
                    .collect();
                if sp_sha1_userids.len() > 0 {
                    certs_sp_sha1_userids += 1;
                }

                if sp_sha1_userids.len() > 0 && sp_userids.len() == 0 {
                    certs_with_only_sha1_protected_userids += 1;
                }

                if sp_userids.len() != sp_sha1_userids.len() {
                    assert!(sp_userids.len() < sp_sha1_userids.len());
                    certs_with_a_sha1_protected_userids += 1;

                    for u in sp_sha1_userids {
                        if ! sp_userids.contains(&u) {
                            println!("Certificate {} contains a user id ({:?}) protected by SHA1",
                                     cert.keyid(),
                                     String::from_utf8_lossy(u.value()));
                        }
                    }
                }


                let sha1_subkeys: Vec<_> = sp_sha1_vc
                    .keys().subkeys()
                    .revoked(false).alive()
                    .for_signing()
                    .collect();
                if sha1_subkeys.len() > 0 {
                    // sp+sha1 is strictly more liberal than sp.
                    // Normally, this means that any keys that are
                    // acceptable under sp+sha1 are acceptable under
                    // sp.  However, if the binding signature uses
                    // SHA256 and a subkey revocation uses SHA1, we
                    // would accept the key under sp+sha1 (it's
                    // revoked), but not sp (it's not revoked)!
                    // Ignore this corner case: we're only gathering
                    // some rough statistics.
                    certs_with_signing_subkeys += 1;

                    let sp_subkeys: Vec<_> = sp_vc
                        .keys().subkeys()
                        .revoked(false).alive()
                        .for_signing()
                        .map(|ka| ka.fingerprint())
                        .collect();

                    // Does this certificate have any subkeys whose
                    // binding signatures use sha1?
                    let mut uses_sha1_somewhere = false;
                    let mut uses_sha1_protected_binding_sig = false;
                    let mut uses_sha1_protected_backsig = false;
                    for ka in sha1_subkeys.into_iter() {
                        if sp_subkeys.contains(&ka.fingerprint()) {
                            continue;
                        }

                        if !uses_sha1_somewhere {
                            certs_with_sha1_protected_signing_subkeys += 1;
                            uses_sha1_somewhere = true;
                        }

                        let sig = ka.binding_signature();

                        if sig.hash_algo() == HashAlgorithm::SHA1 {
                            println!("Certificate {}, key {} uses a \
                                      SHA1-protected binding signature.",
                                     cert.keyid(),
                                     ka.keyid());
                            if ! uses_sha1_protected_binding_sig {
                                uses_sha1_protected_binding_sig = true;
                                certs_with_sha1_protected_binding_sig += 1;
                            }
                            continue;
                        }

                        // The binding signature doesn't use SHA-1.
                        // Then, it must be the backsig.

                        // Is it just the backsig that uses SHA1?
                        let mut backsigs: Vec<_> = sig.embedded_signatures()
                            .filter(|backsig| {
                                (*backsig).clone().verify_primary_key_binding(
                                    &cert.primary_key(),
                                    ka.key()).is_ok()
                            })
                            .collect();
                        if backsigs.len() == 0 {
                            panic!("Valid signing-capable subkey without \
                                    a valid backsig?");
                        }
                        backsigs.sort();
                        backsigs.dedup();

                        if backsigs.len() > 1 {
                            eprintln!("Warning: multiple valid backsigs.");
                        }

                        println!("Certificate {}, key {} uses a \
                                  {}-protected binding signature, but a SHA1-protected backsig",
                                 cert.keyid(),
                                 ka.keyid(),
                                 sig.hash_algo());
                        if ! uses_sha1_protected_backsig {
                            uses_sha1_protected_backsig = true;
                            certs_with_sha1_protected_backsig += 1;
                        }
                    }
                }
            }
            Err(err) => panic!("parsing cert #{}: {}", certs, err),
        }
    }

    println!("{} certificates.", certs);
    println!("{} certificates valid under the standard policy.",
             certs_sp);
    println!("{} certificates valid under the standard policy + sha1.",
             certs_sp_sha1);
    println!("Of the {} certificates valid under the standard policy:",
             certs_sp);
    println!("  {} have >0 user ids under the standard policy",
             certs_sp_userids);
    println!("  {} have >0 user ids under the standard policy + sha1",
             certs_sp_sha1_userids);
    println!("  {} have >0 user ids that are only protected by SHA1",
             certs_with_a_sha1_protected_userids);
    println!("  {} have all user ids only protected by SHA1",
             certs_with_only_sha1_protected_userids);
    println!("Of the {} certificates valid under the standard policy:",
             certs_sp);
    println!("  {} certificates have >0 non-revoked, live, signing-capable subkeys under the standard policy",
             certs_with_signing_subkeys);
    println!("    Of these {} certificates, {} have >0 subkeys protected by SHA1",
             certs_with_signing_subkeys,
             certs_with_sha1_protected_signing_subkeys);
    println!("      Of these {}, {} use SHA1 for the binding signature",
             certs_with_sha1_protected_signing_subkeys,
             certs_with_sha1_protected_binding_sig);
    println!("      Of these {}, {} use something strong for the binding signature, but SHA1 for the backsig",
             certs_with_sha1_protected_signing_subkeys,
             certs_with_sha1_protected_backsig);

    Ok(())
}
